@@ -330,6 +330,11 @@ MASTER_SERV *get_master_ent()
     } else if (STR_SAME(transport, MASTER_XPORT_NAME_FIFO)) {
 	serv->type = MASTER_SERV_TYPE_FIFO;
 	serv->listen_fd_count = 1;
+#ifdef MASTER_SERV_TYPE_PASS
+    } else if (STR_SAME(transport, MASTER_XPORT_NAME_PASS)) {
+	serv->type = MASTER_SERV_TYPE_PASS;
+	serv->listen_fd_count = 1;
+#endif
     } else {
 	fatal_with_context("bad transport type: %s", transport);
     }
@@ -353,6 +358,11 @@ MASTER_SERV *get_master_ent()
     } else if (serv->type == MASTER_SERV_TYPE_FIFO) {
 	serv->name = mail_pathname(private ? MAIL_CLASS_PRIVATE :
 				   MAIL_CLASS_PUBLIC, name);
+#ifdef MASTER_SERV_TYPE_PASS
+    } else if (serv->type == MASTER_SERV_TYPE_PASS) {
+	serv->name = mail_pathname(private ? MAIL_CLASS_PRIVATE :
+				   MAIL_CLASS_PUBLIC, name);
+#endif
     } else {
 	msg_panic("bad transport type: %d", serv->type);
     }
@@ -361,6 +371,10 @@ MASTER_SERV *get_master_ent()
      * Listen socket(s). XXX We pre-allocate storage because the number of
      * sockets is frozen anyway once we build the command-line vector below.
      */
+    if (serv->listen_fd_count == 0) {
+	msg_fatal("%s: line %d: no valid IP address found: %s",
+		  VSTREAM_PATH(master_fp), master_line, name);
+    }
     serv->listen_fd = (int *) mymalloc(sizeof(int) * serv->listen_fd_count);
     for (n = 0; n < serv->listen_fd_count; n++)
 	serv->listen_fd[n] = -1;
@@ -474,6 +488,9 @@ void    print_master_ent(MASTER_SERV *serv)
 	     serv->type == MASTER_SERV_TYPE_UNIX ? MASTER_XPORT_NAME_UNIX :
 	     serv->type == MASTER_SERV_TYPE_FIFO ? MASTER_XPORT_NAME_FIFO :
 	     serv->type == MASTER_SERV_TYPE_INET ? MASTER_XPORT_NAME_INET :
+#ifdef MASTER_SERV_TYPE_PASS
+	     serv->type == MASTER_SERV_TYPE_PASS ? MASTER_XPORT_NAME_PASS :
+#endif
 	     "unknown transport type");
     msg_info("listen_fd_count: %d", serv->listen_fd_count);
     msg_info("wakeup: %d", serv->wakeup_time);
